import builtins
import typing

class FieldIndex:
    """
    *Internal*

    Instances of the `FieldIndex` class represents a field index created by [`vector_index`](#vector-index), [`keyword_index`](#keyword-index), or [`semantic_index`](#semantic-index) functions.
    """
    ...

class FieldSpec:
    """
    *Internal*

    Instances of the `FieldSpec` class represents a field specification created by [`text`](#text), [`int`](#int), [`float`](#float), [`bool`](#bool), [`f32_vector`](#f32-vector), [`u8_vector`](#u8-vector), [`binary_vector`](#binary-vector), [`f32_sparse_vector`](#f32-sparse-vector), [`u8_sparse_vector`](#u8-sparse-vector), [`bytes`](#bytes), or [`list`](#list) functions.
    """

    def required(self) -> FieldSpec: ...
    def optional(self) -> FieldSpec: ...
    def index(self, index: FieldIndex) -> FieldSpec: ...

# data types
def text() -> FieldSpec:
    """
    Create a text field specification.
    """
    ...

def int() -> FieldSpec:
    """
    Create an integer field specification.
    """
    ...

def float() -> FieldSpec:
    """
    Create a float field specification.
    """
    ...

def bool() -> FieldSpec:
    """
    Create a boolean field specification.
    """
    ...

def f32_vector(dimension: builtins.int) -> FieldSpec:
    """
    Create a 32-bit float vector field specification.
    """
    ...

def u8_vector(dimension: builtins.int) -> FieldSpec:
    """
    Create an 8-bit unsigned integer vector field specification.
    """
    ...

def binary_vector(dimension: builtins.int) -> FieldSpec:
    """
    Create a binary vector field specification.
    """
    ...

def f32_sparse_vector() -> FieldSpec:
    """
    Create a f32 sparse vector field specification.
    """
    ...

def u8_sparse_vector() -> FieldSpec:
    """
    Create a u8 sparse vector field specification.
    """
    ...

def bytes() -> FieldSpec:
    """
    Create a bytes field specification.
    """
    ...

def list(value_type: typing.Literal["text", "integer", "float"]) -> FieldSpec:
    """
    Create a list field specification.
    """
    ...

# indexes
def vector_index(
    metric: typing.Literal["cosine", "euclidean", "dot_product", "hamming"],
) -> FieldIndex:
    """
    Create a vector index on a vector field.
    """
    ...

def keyword_index() -> FieldIndex:
    """
    Create a keyword index on a text field.
    """
    ...

def semantic_index(model: str) -> FieldIndex:
    """
    Create a semantic index on a text field.
    """
    ...



import builtins
import typing
from enum import Enum

from . import query, schema


class Client:
    """
    Synchronous client for interacting with the TopK API.
    """

    def __init__(
        self,
        api_key: builtins.str,
        region: builtins.str,
        host: builtins.str = "topk.io",
        https: builtins.bool = True,
        retry_config: typing.Optional[RetryConfig] = None,
    ) -> None:
        ...
    def collection(self, collection: builtins.str) -> CollectionClient:
        """
        Get a client for managing data operations on a specific collection such as querying, upserting, and deleting documents.
        """
        ...
    def collections(self) -> CollectionsClient:
        """
        Get a client for managing collections.
        """
        ...


class AsyncClient:
    """
    Asynchronous client for interacting with the TopK API.
    """

    def __init__(
        self,
        api_key: builtins.str,
        region: builtins.str,
        host: builtins.str = "topk.io",
        https: builtins.bool = True,
        retry_config: typing.Optional[RetryConfig] = None,
    ) -> None:
        ...
    def collection(self, collection: builtins.str) -> AsyncCollectionClient:
        """Get an async client for a specific collection.
        """
        ...
    def collections(self) -> AsyncCollectionsClient:
        """Get an async client for managing collections.
        """
        ...


class CollectionClient:
    """
    Synchronous client for collection operations.
    """

    def get(
        self,
        ids: typing.Sequence[builtins.str],
        fields: typing.Optional[typing.Sequence[builtins.str]] = None,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> builtins.dict[builtins.str, builtins.dict[builtins.str, typing.Any]]:
        """
        Get documents by their IDs.
        """
        ...
    def count(
        self,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> builtins.int:
        """
        Get the count of documents in the collection.
        """
        ...
    def query(
        self,
        query: query.Query,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> builtins.list[builtins.dict[builtins.str, typing.Any]]:
        """
        Execute a query against the collection.
        """
        ...
    def upsert(
        self, documents: typing.Sequence[typing.Mapping[builtins.str, typing.Any]]
    ) -> builtins.str:
        """
        Insert or update documents in the collection.
        """
        ...
    def delete(self, ids: typing.Sequence[builtins.str]) -> builtins.str:
        """
        Delete documents by their IDs.
        """
        ...


class AsyncCollectionClient:
    """
    Asynchronous client for collection operations.
    """

    def get(
        self,
        ids: typing.Sequence[builtins.str],
        fields: typing.Optional[typing.Sequence[builtins.str]] = None,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> typing.Awaitable[builtins.dict[builtins.str, builtins.dict[builtins.str, typing.Any]]]:
        """
        Get documents by their IDs asynchronously.
        """
        ...
    def count(
        self,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> typing.Awaitable[builtins.int]:
        """
        Get the count of documents in the collection asynchronously.
        """
        ...
    def query(
        self,
        query: query.Query,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> typing.Awaitable[builtins.list[builtins.dict[builtins.str, typing.Any]]]:
        """
        Execute a query against the collection asynchronously.
        """
        ...
    def upsert(
        self, documents: typing.Sequence[typing.Mapping[builtins.str, typing.Any]]
    ) -> typing.Awaitable[builtins.str]:
        """
        Insert or update documents in the collection asynchronously.
        """
        ...
    def delete(self, ids: typing.Sequence[builtins.str]) -> typing.Awaitable[builtins.str]:
        """
        Delete documents by their IDs asynchronously.
        """
        ...


class Collection:
    """
    Represents a collection in the TopK system.
    """
    name: builtins.str
    org_id: builtins.str
    project_id: builtins.str
    region: builtins.str
    schema: builtins.dict[builtins.str, schema.FieldSpec]


class CollectionsClient:
    """
    Synchronous client for managing collections.
    """

    def get(self, collection_name: builtins.str) -> Collection:
        """
        Get information about a specific collection.
        """
        ...
    def list(self) -> builtins.list[Collection]:
        """
        List all collections.
        """
        ...
    def create(
        self,
        collection_name: builtins.str,
        schema: typing.Mapping[builtins.str, schema.FieldSpec],
    ) -> Collection:
        """
        Create a new collection with the specified schema.
        """
        ...
    def delete(self, collection_name: builtins.str) -> None:
        """
        Delete a collection.
        """
        ...


class AsyncCollectionsClient:
    """
    Asynchronous client for managing collections.
    """

    def get(self, collection_name: builtins.str) -> typing.Awaitable[Collection]:
        """
        Get information about a specific collection asynchronously.
        """
        ...
    def list(self) -> typing.Awaitable[builtins.list[Collection]]:
        """
        List all collections asynchronously.
        """
        ...
    def create(
        self,
        collection_name: builtins.str,
        schema: typing.Mapping[builtins.str, schema.FieldSpec],
    ) -> typing.Awaitable[Collection]:
        """
        Create a new collection with the specified schema asynchronously.
        """
        ...
    def delete(self, collection_name: builtins.str) -> typing.Awaitable[None]:
        """
        Delete a collection asynchronously.
        """
        ...


class ConsistencyLevel(Enum):
    """
    Enumeration of consistency levels for operations.
    """

    Indexed = "indexed"
    Strong = "strong"


class RetryConfig:
    """
    Configuration for retry behavior.
    """

    max_retries: typing.Optional[builtins.int]
    timeout: typing.Optional[builtins.int]
    backoff: typing.Optional[BackoffConfig]


class BackoffConfig:
    """
    Configuration for backoff behavior in retries.
    """

    base: typing.Optional[builtins.int]
    init_backoff: typing.Optional[builtins.int]
    max_backoff: typing.Optional[builtins.int]



from topk_sdk.data import f32_vector, f64_list, i32_list, i64_list, u8_vector, u32_list
from topk_sdk.query import field, fn, literal, match, select

from . import ProjectContext
from .utils import dataset, doc_ids


def test_query_select_literal(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    results = ctx.client.collection(collection.name).query(
        select(literal=literal(1.0))
        .filter(field("title") == "1984")
        .topk(field("published_year"), 100, True)
    )

    assert results == [{"_id": "1984", "literal": 1.0}]


def test_query_select_non_existing_field(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    results = ctx.client.collection(collection.name).query(
        select(literal=field("non_existing_field"))
        .filter(field("title") == "1984")
        .topk(field("published_year"), 100, True)
    )

    assert results == [{"_id": "1984"}]


def test_query_topk_limit(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    results = ctx.client.collection(collection.name).query(
        select("title").topk(field("published_year"), 3, True)
    )
    assert len(results) == 3

    results = ctx.client.collection(collection.name).query(
        select("title").topk(field("published_year"), 2, True)
    )
    assert len(results) == 2

    results = ctx.client.collection(collection.name).query(
        select("title").topk(field("published_year"), 1, True)
    )
    assert len(results) == 1


def test_query_topk_asc(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    results = ctx.client.collection(collection.name).query(
        select("published_year").topk(field("published_year"), 3, True)
    )

    assert results == [
        {"_id": "pride", "published_year": 1813},
        {"_id": "moby", "published_year": 1851},
        {"_id": "gatsby", "published_year": 1925},
    ]


def test_query_topk_desc(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    results = ctx.client.collection(collection.name).query(
        select("published_year").topk(field("published_year"), 3, False)
    )

    assert results == [
        {"_id": "harry", "published_year": 1997},
        {"_id": "alchemist", "published_year": 1988},
        {"_id": "mockingbird", "published_year": 1960},
    ]


def test_query_select_bm25_score(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    results = ctx.client.collection(collection.name).query(
        select(bm25_score=fn.bm25_score())
        .filter(match("pride"))
        .topk(field("bm25_score"), 100, True)
    )

    assert doc_ids(results) == {"pride"}


def test_query_select_vector_distance(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    results = ctx.client.collection(collection.name).query(
        select(
            summary_distance=fn.vector_distance("summary_embedding", [2.0] * 16)
        ).topk(field("summary_distance"), 3, True)
    )

    assert doc_ids(results) == {"1984", "mockingbird", "pride"}


def test_query_select_null_field(ctx: ProjectContext):
    collection = ctx.client.collections().create(ctx.scope("test"), schema={})

    ctx.client.collection(collection.name).upsert(
        [{"_id": "1984", "a": None}, {"_id": "pride"}]
    )

    results = ctx.client.collection(collection.name).query(
        select(a=field("a"), b=literal(1)).topk(field("b"), 100, True)
    )

    # Assert that `a` is null for all documents, even when not specified when upserting
    assert {doc.get("a") for doc in results} == {None, None}


def test_query_select_text_match(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    results = ctx.client.collection(collection.name).query(
        select(
            match_surveillance=field("summary").match_all("surveillance control mind"),
            match_love=field("summary").match_any("love class marriage"),
        )
        .filter((field("title") == "1984") | (field("_id") == "pride"))
        .topk(field("published_year"), 100, True)
    )

    assert results == [
        {"_id": "pride", "match_surveillance": False, "match_love": True},
        {"_id": "1984", "match_surveillance": True, "match_love": False},
    ]


def test_query_select_union(ctx: ProjectContext):
    # create collection
    collection = ctx.client.collections().create(ctx.scope("test"), schema={})

    # upsert documents with different types in the same field
    lsn = ctx.client.collection(collection.name).upsert(
        [
            {"_id": "0", "rank": 0, "mixed": None},
            {"_id": "1", "rank": 1, "mixed": 1},
            {"_id": "2", "rank": 2, "mixed": 2},
            {"_id": "3", "rank": 3, "mixed": 3},
            {"_id": "4", "rank": 4, "mixed": 4},
            {"_id": "5", "rank": 5, "mixed": 5.0},
            {"_id": "6", "rank": 6, "mixed": 6.0},
            {"_id": "7", "rank": 7, "mixed": True},
            {"_id": "8", "rank": 8, "mixed": "hello"},
            {"_id": "9", "rank": 9, "mixed": u8_vector([1, 2, 3])},
            {"_id": "10", "rank": 10, "mixed": f32_vector([1.0, 2.0, 3.0])},
            {"_id": "11", "rank": 11, "mixed": bytes([1, 2, 3])},
            {"_id": "12", "rank": 12, "mixed": u32_list([17, 6, 1997])},
            {"_id": "13", "rank": 13, "mixed": i32_list([17, 6, 1997])},
            {"_id": "14", "rank": 14, "mixed": i64_list([17, 6, 1997])},
            {"_id": "15", "rank": 15, "mixed": f64_list([17.5, 6.5, 1997.5])},
            {"_id": "16", "rank": 16, "mixed": ["foo", "bar"]},
        ]
    )

    # wait for writes to be flushed
    ctx.client.collection(collection.name).count(lsn=lsn)

    results = ctx.client.collection(collection.name).query(
        select("mixed").topk(field("rank"), 100, True)
    )

    # Verify we have all the documents
    assert results == [
        {"_id": "0", "mixed": None},
        {"_id": "1", "mixed": 1},
        {"_id": "2", "mixed": 2},
        {"_id": "3", "mixed": 3},
        {"_id": "4", "mixed": 4},
        {"_id": "5", "mixed": 5.0},
        {"_id": "6", "mixed": 6.0},
        {"_id": "7", "mixed": True},
        {"_id": "8", "mixed": "hello"},
        {"_id": "9", "mixed": [1, 2, 3]},
        {"_id": "10", "mixed": [1.0, 2.0, 3.0]},
        {"_id": "11", "mixed": bytes([1, 2, 3])},
        {"_id": "12", "mixed": [17, 6, 1997]},
        {"_id": "13", "mixed": [17, 6, 1997]},
        {"_id": "14", "mixed": [17, 6, 1997]},
        {"_id": "15", "mixed": [17.5, 6.5, 1997.5]},
        {"_id": "16", "mixed": ["foo", "bar"]},
    ]



class CollectionAlreadyExistsError(Exception):
  """
  Raised when creating a collection with a name that already exists.
  """


class CollectionNotFoundError(Exception):
  """
  Raised when a collection is not found.
  """


class CollectionValidationError(Exception):
  """
  Raised when a collection name or schema is invalid.
  """


class DocumentValidationError(Exception):
  """
  Raised when a document is invalid.
  """


class InvalidArgumentError(Exception):
  """
  Raised when an invalid argument is provided.
  """


class PermissionDeniedError(Exception):
  """
  Raised when a permission is denied.
  """


class QueryLsnTimeoutError(Exception):
  """
  Raised when a query LSN timeout occurs.
  """


class QuotaExceededError(Exception):
  """
  Raised when a quota is exceeded.
  """


class RequestTooLargeError(Exception):
  """
  Raised when a request is too large.
  """


class SchemaValidationError(Exception):
  """
  Raised when a schema is invalid.
  """


class SlowDownError(Exception):
  """
  Raised when a slow down occurs.
  """



#!/usr/bin/env python3
"""Main script for generating API documentation from Python type stubs."""

import sys
from pathlib import Path

from parse import parse_module
from render import render_module


def generate_docs_for_file(input_file: Path, output_dir: Path, base_dir: Path):
    """Generate documentation for a single .pyi file."""
    # Calculate relative path from base directory
    rel_path = input_file.relative_to(base_dir)

    # Create output file path (replace .pyi with .md)
    # Use the parent folder name as the file name
    parent_folder_name = rel_path.parent.name if rel_path.parent.name else rel_path.stem
    output_file = output_dir / f"{parent_folder_name}.mdx"

    if rel_path == Path("__init__.pyi"):
        output_file = output_dir / "index.mdx"

    # Create output directory if it doesn't exist
    output_file.parent.mkdir(parents=True, exist_ok=True)

    # Parse the module
    module = parse_module(input_file)

    # Render to file instead of stdout
    with open(output_file, "w") as f:
        import sys
        from io import StringIO

        # Capture the output
        old_stdout = sys.stdout
        sys.stdout = StringIO()

        try:
            render_module(module)
            output = sys.stdout.getvalue()
        finally:
            sys.stdout = old_stdout

        f.write(output)

    print(f"Generated: {output_file}")


def main():
    """Main entry point."""
    if len(sys.argv) < 2 or len(sys.argv) > 3:
        print("Usage: python main.py <path_to_pyi_files> [output_directory]")
        print(
            "  path_to_pyi_files: Path to .pyi file or directory containing .pyi files"
        )
        print("  output_directory: Optional output directory (default: 'out')")
        sys.exit(1)

    input_path = Path(sys.argv[1])
    output_dir = Path(sys.argv[2]) if len(sys.argv) == 3 else Path("docs")

    if not input_path.exists():
        print(f"Error: Path {input_path} does not exist")
        sys.exit(1)

    # Find all .pyi files
    if input_path.is_file() and input_path.suffix == ".pyi":
        # Single file
        files = [input_path]
        base_dir = input_path.parent
    else:
        # Directory - find all .pyi files
        files = list(input_path.rglob("**/*.pyi"))
        base_dir = input_path

    if not files:
        print("No .pyi files found")
        sys.exit(1)

    print(f"Found {len(files)} .pyi files")
    print(f"Output directory: {output_dir}")
    print(f"Base directory: {base_dir}")
    print()

    # Generate documentation for each file
    for file in files:
        try:
            generate_docs_for_file(file, output_dir, base_dir)
        except Exception as e:
            print(f"Error processing {file}: {e}")
            continue

    print(f"\nDocumentation generation complete! Check the '{output_dir}' directory.")


if __name__ == "__main__":
    main()



import builtins
import typing

class List:
    """
    *Internal*

    Instances of the `List` class are used to represent lists of values in TopK.
    Usually created using data constructors such as [`f32_list()`](#f32-list), [`i32_list()`](#i32-list), etc.
    """
    ...


class SparseVector:
    """
    *Internal*

    Instances of the `SparseVector` class are used to represent sparse vectors in TopK.
    Usually created using data constructors such as [`f32_sparse_vector()`](#f32-sparse-vector) or [`u8_sparse_vector()`](#u8-sparse-vector).
    """
    ...


def f32_vector(data: builtins.list[float]) -> List:
    """
    Create a 32-bit float vector.
    """
    ...


def u8_vector(data: builtins.list[int]) -> List:
    """
    Create an 8-bit unsigned integer vector.
    """
    ...


def binary_vector(data: builtins.list[int]) -> List:
    """
    Create a binary vector.
    """
    ...


def f32_sparse_vector(data: builtins.dict[int, float]) -> SparseVector:
    """
    Create a 32-bit float sparse vector.
    """
    ...


def u8_sparse_vector(data: builtins.dict[int, int]) -> SparseVector:
    """
    Create an 8-bit unsigned integer sparse vector.
    """
    ...


def bytes(data: typing.Union[builtins.list[int], builtins.bytes]) -> List:
    """
    Create a bytes data structure.
    """
    ...


def u32_list(data: builtins.list[int]) -> List:
    """
    Create a list of 32-bit unsigned integers.
    """
    ...


def i32_list(data: builtins.list[int]) -> List:
    """
    Create a list of 32-bit signed integers.
    """
    ...


def i64_list(data: builtins.list[int]) -> List:
    """
    Create a list of 64-bit signed integers.
    """
    ...


def f32_list(data: builtins.list[float]) -> List:
    """
    Create a list of 32-bit floating point numbers.
    """
    ...


def f64_list(data: builtins.list[float]) -> List:
    """
    Create a list of 64-bit floating point numbers.
    """
    ...


def string_list(data: builtins.list[str]) -> List:
    """
    Create a list of strings.
    """
    ...



import builtins
import typing
from enum import Enum

import topk_sdk.data

class LogicalExpr(Enum):
    """
    *Internal*

    Instances of the `LogicalExpr` class are used to represent logical expressions in TopK.
    Usually created using logical constructors such as [`field()`](#field), [`literal()`](#literal), etc.
    """

    def __repr__(self) -> builtins.str:
        ...
    def _expr_eq(self, other: LogicalExpr) -> LogicalExpr:
        ...
    def is_null(self) -> LogicalExpr:
        """
        Check if the expression is null.
        """
        ...
    def is_not_null(self) -> LogicalExpr:
        """
        Check if the expression is not null.
        """
        ...
    def abs(self) -> LogicalExpr:
        """
        Compute the absolute value of the expression.
        """
        ...
    def __abs__(self) -> LogicalExpr:
        """
        Compute the absolute value of the expression.
        """
        ...
    def ln(self) -> LogicalExpr:
        """
        Compute the natural logarithm of the expression.
        """
        ...
    def exp(self) -> LogicalExpr:
        """
        Compute the exponential of the expression.
        """
        ...
    def sqrt(self) -> LogicalExpr:
        """
        Compute the square root of the expression.
        """
        ...
    def square(self) -> LogicalExpr:
        """
        Compute the square of the expression.
        """
        ...
    def eq(self, other: FlexibleExpr) -> LogicalExpr:
        """
        Check if the expression is equal to another expression.
        """
        ...
    def __eq__(self, other: FlexibleExpr) -> LogicalExpr:
        """
        Check if the expression is equal to another expression using the `==` operator.
        """
        ...
    def ne(self, other: FlexibleExpr) -> LogicalExpr:
        """
        Check if the expression is not equal to another expression.
        """
        ...
    def __ne__(self, other: FlexibleExpr) -> LogicalExpr:
        """
        Check if the expression is not equal to another expression using the `!=` operator.
        """
        ...
    def lt(self, other: Ordered) -> LogicalExpr:
        """
        Check if the expression is less than another expression.
        """
        ...
    def __lt__(self, other: Ordered) -> LogicalExpr:
        """
        Check if the expression is less than another expression using the `<` operator.
        """
        ...
    def __rlt__(self, other: Ordered) -> LogicalExpr:
        """
        Check if the other expression is less than the self expression using the `>` operator.
        """
        ...
    def lte(self, other: Ordered) -> LogicalExpr:
        """
        Check if the expression is less than or equal to another expression.
        """
        ...
    def __le__(self, other: Ordered) -> LogicalExpr:
        """
        Check if the expression is less than or equal to another expression using the `<=` operator.
        """
        ...
    def __rle__(self, other: Ordered) -> LogicalExpr:
        """
        Check if the other expression is less than or equal to the self expression using the `<=` operator.
        """
        ...
    def gt(self, other: Ordered) -> LogicalExpr:
        """
        Check if the expression is greater than another expression.
        """
        ...
    def __gt__(self, other: Ordered) -> LogicalExpr:
        """
        Check if the expression is greater than another expression using the `>` operator.
        """
        ...
    def __rgt__(self, other: Ordered) -> LogicalExpr:
        """
        Check if the expression is greater than another expression using the `<` operator.
        """
        ...
    def gte(self, other: Ordered) -> LogicalExpr:
        """
        Check if the expression is greater than or equal to another expression.
        """
        ...
    def __ge__(self, other: Ordered) -> LogicalExpr:
        """
        Check if the expression is greater than or equal to another expression using the `>=` operator.
        """
        ...
    def __rge__(self, other: Ordered) -> LogicalExpr:
        """
        Check if the other expression is greater than or equal to the self expression using the `<=` operator.
        """
        ...
    def add(self, other: Numeric) -> LogicalExpr:
        """
        Add another value to the expression.
        """
        ...
    def __add__(self, other: Numeric) -> LogicalExpr:
        """
        Add another value to the expression using the `+` operator.
        """
        ...
    def __radd__(self, other: Numeric) -> LogicalExpr:
        """
        Add the other value to the self expression using the `+` operator.
        """
        ...
    def sub(self, other: Numeric) -> LogicalExpr:
        """
        Subtract another value from the expression.
        """
        ...
    def __sub__(self, other: Numeric) -> LogicalExpr:
        """
        Subtract another value from the expression using the `-` operator.
        """
        ...
    def __rsub__(self, other: Numeric) -> LogicalExpr:
        """
        Subtract the other value from the self expression using the `-` operator.
        """
        ...
    def mul(self, other: Numeric) -> LogicalExpr:
        """
        Multiply the expression by another value.
        """
        ...
    def __mul__(self, other: Numeric) -> LogicalExpr:
        """
        Multiply the expression by another value using the `*` operator.
        """
        ...
    def __rmul__(self, other: Numeric) -> LogicalExpr:
        """
        Multiply the other value by the self expression using the `*` operator.
        """
        ...
    def div(self, other: Numeric) -> LogicalExpr:
        """
        Divide the expression by another value.
        """
        ...
    def __div__(self, other: Numeric) -> LogicalExpr:
        """
        Divide the expression by another value using the `/` operator.
        """
        ...
    def __truediv__(self, other: Numeric) -> LogicalExpr:
        """
        Divide the expression by another value using the `/` operator.
        """
        ...
    def __rdiv__(self, other: Numeric) -> LogicalExpr:
        """
        Divide the other value by the self expression using the `/` operator.
        """
        ...
    def __rtruediv__(self, other: Numeric) -> LogicalExpr:
        """
        Divide the other value by the self expression using the `/` operator.
        """
        ...
    def min(self, other: Ordered) -> LogicalExpr:
        """
        Compute the minimum of the expression and another value.
        """
        ...
    def max(self, other: Ordered) -> LogicalExpr:
        """
        Compute the maximum of the expression and another value.
        """
        ...
    def and_(self, other: Boolish) -> LogicalExpr:
        """
        Compute the logical AND of the expression and another expression.
        """
        ...
    def __and__(self, other: Boolish) -> LogicalExpr:
        """
        Compute the logical AND of the expression and another expression using the `&` operator.
        """
        ...
    def __rand__(self, other: Boolish) -> LogicalExpr:
        """
        Compute the logical AND of the other expression and the self expression using the `&` operator.
        """
        ...
    def or_(self, other: Boolish) -> LogicalExpr:
        """
        Compute the logical OR of the expression and another expression.
        """
        ...
    def __or__(self, other: Boolish) -> LogicalExpr:
        """
        Compute the logical OR of the expression and another expression using the `|` operator.
        """
        ...
    def __ror__(self, other: Boolish) -> LogicalExpr:
        """
        Compute the logical OR of the other expression and the self expression using the `|` operator.
        """
        ...
    def starts_with(self, other: Stringy) -> LogicalExpr:
        """
        Check if the expression starts with the provided string expression.
        """
        ...
    def contains(self, other: FlexibleExpr) -> LogicalExpr:
        """
        Check if the expression contains another value.
        """
        ...
    def in_(self, other: Iterable) -> LogicalExpr:
        """
        Check if the expression is in the provided iterable expression.
        """
        ...
    def match_all(self, other: StringyWithList) -> LogicalExpr:
        """
        Check if the expression matches all terms against the field with keyword index.
        """
        ...
    def match_any(self, other: StringyWithList) -> LogicalExpr:
        """
        Check if the expression matches any term against the field with keyword index.
        """
        ...
    def coalesce(self, other: Numeric) -> LogicalExpr:
        """
        Coalesce nulls in the expression with another value.
        """
        ...
    def choose(self, x: FlexibleExpr, y: FlexibleExpr) -> LogicalExpr:
        """
        Choose between two values based on the expression.
        """
        ...
    def boost(self, condition: FlexibleExpr, boost: Numeric) -> LogicalExpr:
        """
        Multiply the scoring expression by the provided `boost` value if the `condition` is true.
        """
        ...

FlexibleExpr = typing.Union[str, int, float, bool, None, LogicalExpr]
Numeric = typing.Union[int, float, LogicalExpr]
Ordered = typing.Union[int, float, str, LogicalExpr]
Boolish = typing.Union[bool, LogicalExpr]
Stringy = typing.Union[str, LogicalExpr]
StringyWithList = typing.Union[str, builtins.list[str], LogicalExpr]
Iterable = typing.Union[str, builtins.list[int], builtins.list[float], builtins.list[str], topk_sdk.data.List, LogicalExpr]

class FunctionExpr:
    """
    *Internal*

    Instances of the `FunctionExpr` class are used to represent function expressions in TopK.
    Usually created using function constructors such as [`fn.vector_distance()`](#vector-distance), [`fn.semantic_similarity()`](#semantic-similarity) or [`fn.bm25_score()`](#bm25-score).
    """
    ...

class TextExpr(Enum):
    """
    *Internal*

    Instances of the `TextExpr` class are used to represent text expressions in TopK.
    """

    def __and__(self, other: TextExpr) -> TextExpr:
        """
        Combine the expression with another text expression using the `&` operator.
        """
        ...
    def __rand__(self, other: TextExpr) -> TextExpr:
        """
        Combine the other text expression with the self expression using the `&` operator.
        """
        ...
    def __or__(self, other: TextExpr) -> TextExpr:
        """
        Combine the expression with another text expression using the `|` operator.
        """
        ...
    def __ror__(self, other: TextExpr) -> TextExpr:
        """
        Combine the other text expression with the self expression using the `|` operator.
        """
        ...


class Query:
    def select(
        self,
        *args: builtins.str,
        **kwargs: typing.Union[LogicalExpr, FunctionExpr],
    ) -> Query: ...
    def filter(self, expr: LogicalExpr | TextExpr) -> Query: ...
    def topk(
        self, expr: LogicalExpr, k: builtins.int, asc: builtins.bool = False
    ) -> Query: ...
    def rerank(
        self,
        model: typing.Optional[builtins.str] = None,
        query: typing.Optional[builtins.str] = None,
        fields: typing.Sequence[builtins.str] = [],
        topk_multiple: typing.Optional[builtins.int] = None,
    ) -> Query: ...
    def count(self) -> Query: ...


def field(name: builtins.str) -> LogicalExpr:
    """
    Select a field from the document.
    """
    ...


def select(
    *args: builtins.str,
    **kwargs: typing.Union[LogicalExpr, FunctionExpr],
) -> Query:
    """
    # Example:

    Create a select stage of a query.

    ```python
    # Example:

    from topk_sdk.query import select, field

    client.collection("books").query(
      select("title", year=field("published_year"))
    )
    ```
    """
...


def filter(expr: LogicalExpr | TextExpr) -> Query:
    """
    Create a filter stage of a query.

    ```python
    # Example:

    from topk_sdk.query import filter, field

    client.collection("books").query(
      filter(field("published_year") > 1980)
    )
    ```
    """
    ...


def literal(value: typing.Any) -> LogicalExpr:
    """
    Create a literal expression.
    """
    ...


def match(
    token: builtins.str,
    field: builtins.str | None = None,
    weight: builtins.float = 1.0,
    all: builtins.bool = False,
) -> LogicalExpr:
    """
    Perform a keyword search for documents that contain specific keywords or phrases.

    This function should be used in the filter stage of a query. You can configure
    the match() function to:
    - Match on multiple terms
    - Match only on specific fields
    - Use weights to prioritize certain terms
    """
...


def not_(expr: LogicalExpr) -> LogicalExpr:
    """
    Negate a logical expression.

    ```python
    # Example:

    from topk_sdk.query import field, not_

    .filter(
        not_(field("title").contains("Catcher"))
    )
    ```
    """
    ...


def abs(expr: LogicalExpr) -> LogicalExpr:
    """
    Compute the absolute value of a logical expression.

    ```python
    # Example:

    from topk_sdk.query import field, abs

    client.collection("books").query(
      filter(abs(field("rating")) > 4.5)
    )
    ```
    """


def all(exprs: typing.Sequence[LogicalExpr]) -> LogicalExpr:
    """
    Create a logical AND expression.

    ```python
    # Example:

    from topk_sdk.query import field, all

    client.collection("books").query(
      filter(all([
        field("published_year") >= 1900,
        field("published_year") <= 2000,
        field("title").is_not_null()
      ]))
    )
    ```
    """


def any(exprs: typing.Sequence[LogicalExpr]) -> LogicalExpr:
    """
    Create a logical OR expression.

    ```python
    # Example:

    from topk_sdk.query import field, any

    client.collection("books").query(
      filter(any([
        field("genre") == "fiction",
        field("genre") == "mystery",
        field("genre") == "thriller"
      ]))
    )
    ```
    """
    ...


def min(left: Ordered, right: Ordered) -> LogicalExpr:
    """
    Create a logical MIN expression.

    ```python
    # Example:

    from topk_sdk.query import field, min

    client.collection("books").query(
      filter(min(field("rating"), field("published_year")))
    )
    ```
    """
    ...


def max(left: Ordered, right: Ordered) -> LogicalExpr:
    """
    Create a logical MAX expression.

    ```python
    from topk_sdk.query import field, max

    client.collection("books").query(
      filter(max(field("rating"), field("published_year")))
    )
    ```
    """
    ...


class fn:
    """
    The `query.fn` submodule exposes functions for creating function expressions such as [`fn.vector_distance()`](#vector-distance), [`fn.semantic_similarity()`](#semantic-similarity) or [`fn.bm25_score()`](#bm25-score).
    """
    ...

    @staticmethod
    def vector_distance(
        field: builtins.str,
        vector: typing.Union[
            list[int],
            list[float],
            dict[int, float],
            dict[int, int],
            topk_sdk.data.SparseVector,
            topk_sdk.data.List,
        ],
        skip_refine: builtins.bool = False,
    ) -> FunctionExpr:
        """
        Calculate the vector distance between a field and a query vector.

        ```python
        # Example:

        from topk_sdk.query import field, fn, select

        client.collection("books").query(
          select(
            "title",
            title_similarity=fn.vector_distance(
              "title_embedding",
              [0.1, 0.2, 0.3, ...] # embedding for "animal"
            )
          )
          .topk(field("title_similarity"), 10)
        )
        ```
        """
        ...
    @staticmethod
    def semantic_similarity(
        field: builtins.str,
        query: builtins.str,
    ) -> FunctionExpr:
        """
        Calculate the semantic similarity between a field and a query string.

        ```python
        # Example:

        from topk_sdk.query import field, fn, select

        client.collection("books").query(
          select(
            "title",
            title_similarity=fn.semantic_similarity("title", "animal")
          )
          .topk(field("title_similarity"), 10)
        )
        ```
        """
        ...
    @staticmethod
    def bm25_score() -> FunctionExpr:
        """
        Calculate the BM25 score for a keyword search.

        ```python
        # Example:

        from topk_sdk.query import field, fn, select

        client.collection("books").query(
          select(
            "title",
            text_score=fn.bm25_score()
          )
          .filter(match("animal"))
          .topk(field("text_score"), 10)
        )
        ```
        """
        ...



"""Renderer module for converting parsed data structures to markdown."""

from typing import List

from parse import (
    Class,
    EnumValue,
    Function,
    Method,
    Module,
    Parameter,
    Property,
    TypeAlias,
    TypeAnnotation,
)


def format_type_annotation_str(
    type_annotation: TypeAnnotation, with_links: bool = True
) -> str:
    """Format a TypeAnnotation to a string."""
    if type_annotation.is_generic:
        if type_annotation.generic_args:
            if type_annotation.name == "Union":
                # Format union types as A | B instead of Union[A, B]
                args_str = " | ".join(
                    format_type_annotation_str(arg, with_links)
                    for arg in type_annotation.generic_args
                )
                return args_str
            else:
                args_str = ", ".join(
                    format_type_annotation_str(arg, with_links)
                    for arg in type_annotation.generic_args
                )
                base = f"{type_annotation.name}[{args_str}]"
        else:
            base = type_annotation.name
    else:
        base = type_annotation.name

    if with_links and should_link_type(base):
        return create_type_link(base)
    else:
        return base


def should_link_type(type_str: str) -> bool:
    """Check if a type should be linked."""
    # Skip generic types
    if any(
        generic in type_str
        for generic in [
            "[",
            "]",
            "|",
            "Union",
            "Optional",
            "Sequence",
            "Mapping",
            "Awaitable",
        ]
    ):
        return False

    # Skip built-in types
    builtin_types = {
        "str",
        "int",
        "float",
        "bool",
        "None",
        "Any",
        "dict",
        "list",
        "tuple",
        "set",
    }
    if type_str in builtin_types:
        return False

    # Link topk_sdk module references (both full and shortened paths)
    if type_str.startswith("topk_sdk."):
        return True

    # Link shortened module references (data., query., schema., error.)
    if any(type_str.startswith(prefix) for prefix in ["data.", "query.", "schema.", "error."]):
        return True

    # Only link if it looks like a class name (starts with capital letter)
    return bool(type_str and type_str[0].isupper())


def create_type_link(type_str: str) -> str:
    """Create a markdown link for a type."""
    # Handle cross-module references
    if "." in type_str:
        parts = type_str.split(".")

        # Handle full module path: topk_sdk.module.Class
        if type_str.startswith("topk_sdk.") and len(parts) >= 3:
            module_name = parts[1]  # e.g., "data"
            class_name = parts[-1]  # e.g., "SparseVector"
            return f"[`{type_str}`](/sdk/topk-py/{module_name}#{class_name.lower()})"

        # Handle shortened module path: module.Class (data.SparseVector, query.Expr, etc.)
        elif len(parts) == 2 and parts[0] in ["data", "query", "schema", "error"]:
            module_name = parts[0]  # e.g., "data"
            class_name = parts[1]  # e.g., "SparseVector"
            return f"[`{type_str}`](/sdk/topk-py/{module_name}#{class_name.lower()})"

        # Handle main module: topk_sdk.Class
        elif type_str.startswith("topk_sdk.") and len(parts) == 2:
            class_name = parts[1]
            return f"[`{type_str}`](/sdk/topk-py/index#{class_name.lower()})"

    # For local references, create anchor link
    link = type_str.replace(".", "#").lower()
    return f"[`{type_str}`](#{link})"


def format_function_signature(
    method: Method, class_name: str = "", with_links: bool = True
) -> str:
    """Format a method signature."""
    args: List[str] = []
    for param in method.parameters:
        arg_str = param.name
        if param.type_annotation:
            type_str = format_type_annotation_str(param.type_annotation, with_links)
            arg_str += f": {type_str}"
        if param.default_value:
            arg_str += f" = {param.default_value}"
        args.append(arg_str)

    # For constructors, format more nicely
    if method.is_constructor:
        # Remove 'self' from constructor signature
        args = args[1:] if args and args[0].startswith("self") else args
        signature = f"{class_name or 'Class'}({', '.join(args)})"
    else:
        signature = f"{method.name}({', '.join(args)})"

    if method.return_type:
        return_type_str = format_type_annotation_str(method.return_type, with_links)
        signature += f" -> {return_type_str}"

    # Format as multi-line if signature is long
    if len(signature) > 100 and len(args) > 2:
        if method.is_constructor:
            base = f"{class_name or 'Class'}("
        else:
            base = f"{method.name}("

        lines = [base]
        for i, arg in enumerate(args):
            if i == 0:
                lines.append(f"   {arg},")
            elif i == len(args) - 1:
                lines.append(f"   {arg}")
            else:
                lines.append(f"   {arg},")
        lines.append(")")
        signature = "\n".join(lines)

    return signature


def render_enum_values(enum_values: List[EnumValue]):
    """Render enum values as a table."""
    print("**Values**")
    print()
    print("| Value | Description |")
    print("| ----- | ----------- |")
    for value in enum_values:
        print(f"| `{value.name}` | `{value.value}` |")


def render_data_class_properties(properties: List[Property]):
    """Render data class properties as a table."""
    print("**Properties**")
    print()
    print("| Property | Type |")
    print("| -------- | ---- |")
    for prop in properties:
        type_str = format_type_annotation_str(prop.type_annotation, with_links=True)
        print(f"| `{prop.name}` | {type_str} |")


def render_constructor(method: Method, class_name: str):
    """Render constructor method."""
    print("**Constructor**")
    print()
    print("```python")
    signature = format_function_signature(method, class_name, with_links=False)
    print(signature)
    print("```")
    print()
    if method.docstring:
        print(method.docstring)
        print()


def render_method(method: Method):
    """Render a regular method."""

    escaped_name = method.name.replace('_', r'\_')
    print(f"#### {escaped_name}()")
    print()

    print("```python")
    signature = format_function_signature(method, with_links=False)
    print(signature)
    print("```")
    print()

    if method.docstring:
        print(method.docstring)
        print()


    # Add Parameters section
    if method.parameters and not (
        len(method.parameters) == 1 and method.parameters[0].name == "self"
    ):
        print("**Parameters**")
        print()
        print("| Parameter | Type |")
        print("| --------- | ---- |")
        for param in method.parameters:
            if param.name == "self":
                continue
            type_str = (
                format_type_annotation_str(param.type_annotation, with_links=True)
                if param.type_annotation
                else "Any"
            )
            print(f"| `{param.name}` | {type_str} |")
        print()

    # Add Returns section
    if method.return_type:
        return_type_str = format_type_annotation_str(
            method.return_type, with_links=True
        )
        print("**Returns**")
        print()
        print(return_type_str)
        print()

    print("***")
    print()


def render_class_methods(methods: List[Method], class_name: str):
    """Render all methods in a class."""
    print("**Methods**")
    print()
    for method in methods:
        # Skip __new__ methods for data classes
        if method.name == "__new__" and len(methods) == 1:
            continue

        if method.is_constructor:
            render_constructor(method, class_name)
        else:
            render_method(method)


def render_class(cls: Class, file_path):
    """Render a complete class."""
    print(f"### {cls.name}")
    print()

    if cls.docstring:
        print(cls.docstring)
        print()

    if cls.is_enum and cls.enum_values:
        render_enum_values(cls.enum_values)
    elif cls.is_data_class and cls.properties:
        render_data_class_properties(cls.properties)
    elif cls.methods:
        render_class_methods(cls.methods, cls.name)


def render_function(func: Function, file_path):
    """Render a standalone function."""
    print(f"### {func.name}()")
    print()

    print("```python")
    # Create a temporary method object for formatting
    temp_method = Method(
        name=func.name,
        parameters=func.parameters,
        return_type=func.return_type,
        docstring=func.docstring,
    )
    signature = format_function_signature(temp_method, with_links=False)
    print(signature)
    print("```")
    print()

    if func.docstring:
        print(func.docstring)
        print()

    # Add Parameters section
    if func.parameters:
        print("**Parameters**")
        print()
        print("| Parameter | Type |")
        print("| --------- | ---- |")
        for param in func.parameters:
            type_str = (
                format_type_annotation_str(param.type_annotation, with_links=True)
                if param.type_annotation
                else "Any"
            )
            print(f"| `{param.name}` | {type_str} |")
        print()

    # Add Returns section
    if func.return_type:
        return_type_str = format_type_annotation_str(func.return_type, with_links=True)
        print("**Returns**")
        print()
        print(return_type_str)
        print()

    print("***")
    print()


def render_type_alias(type_alias: TypeAlias, file_path):
    """Render a type alias."""
    print(f"### {type_alias.name}")
    print()

    if type_alias.docstring:
        print(type_alias.docstring)
        print()

    print("```python")
    type_str = format_type_annotation_str(type_alias.type_annotation, with_links=False)
    print(f"{type_alias.name} = {type_str}")
    print("```")
    print()

    print("**Type**")
    print()
    type_str_with_links = format_type_annotation_str(type_alias.type_annotation, with_links=True)
    print(type_str_with_links)
    print()

    print("***")
    print()


def render_module(module: Module):
    """Render a complete module."""
    print("---")

    if module.file_path.parent.name == "topk_sdk":
        print(f"title: topk_sdk")
    else:
        print(f"title: topk_sdk.{module.file_path.parent.name}")

    print("---")
    print()

    print("## Classes")
    print()

    for cls in module.classes:
        render_class(cls, module.file_path)

    if (module.functions.__len__() > 0):
        print("## Functions")
        print()

        for func in module.functions:
            render_function(func, module.file_path)

    if (module.type_aliases and module.type_aliases.__len__() > 0):
        print("## Type Aliases")
        print()

        for type_alias in module.type_aliases:
            render_type_alias(type_alias, module.file_path)



import pytest
from topk_sdk import data, error, query, schema

from . import ProjectContext


def test_upsert_to_non_existent_collection(ctx: ProjectContext):
    with pytest.raises(error.CollectionNotFoundError):
        ctx.client.collection("missing").upsert([{"_id": "one"}])


def test_upsert_basic(ctx: ProjectContext):
    collection = ctx.client.collections().create(ctx.scope("test"), schema={})

    lsn = ctx.client.collection(collection.name).upsert([{"_id": "one"}])
    assert lsn == "1"


def test_upsert_batch(ctx: ProjectContext):
    collection = ctx.client.collections().create(ctx.scope("test"), schema={})

    lsn = ctx.client.collection(collection.name).upsert(
        [{"_id": "one"}, {"_id": "two"}]
    )
    assert lsn == "1"


def test_upsert_sequential(ctx: ProjectContext):
    collection = ctx.client.collections().create(ctx.scope("test"), schema={})

    lsn = ctx.client.collection(collection.name).upsert([{"_id": "one"}])
    assert lsn == "1"

    lsn = ctx.client.collection(collection.name).upsert([{"_id": "two"}])
    assert lsn == "2"

    lsn = ctx.client.collection(collection.name).upsert([{"_id": "three"}])
    assert lsn == "3"


def test_upsert_no_documents(ctx: ProjectContext):
    collection = ctx.client.collections().create(ctx.scope("test"), schema={})

    with pytest.raises(error.DocumentValidationError) as exc_info:
        ctx.client.collection(collection.name).upsert([])
    assert "NoDocuments" in str(exc_info.value)


def test_upsert_invalid_document(ctx: ProjectContext):
    collection = ctx.client.collections().create(ctx.scope("test"), schema={})

    with pytest.raises(error.DocumentValidationError) as exc_info:
        ctx.client.collection(collection.name).upsert([{}])
    assert "MissingId" in str(exc_info.value)


def test_upsert_schema_validation(ctx: ProjectContext):
    collection = ctx.client.collections().create(
        ctx.scope("test"), schema={"name": schema.text().required()}
    )

    with pytest.raises(error.DocumentValidationError) as exc_info:
        ctx.client.collection(collection.name).upsert([{"_id": "one"}])
    assert "MissingField" in str(exc_info.value)


def test_upsert_max_doc_size(ctx: ProjectContext):
    collection = ctx.client.collections().create(ctx.scope("test"), schema={})

    with pytest.raises(error.DocumentValidationError) as exc_info:
        ctx.client.collection(collection.name).upsert(
            [
                {"_id": "one", "payload": "x" * (500 * 1024)},  # 500KB, too large
                {"_id": "two", "payload": "xxx"},               # ok
                {"_id": "three", "payload": "x" * (130 * 1024)},  # 130KB, too large
                {"_id": "four", "payload": "x" * (126 * 1024)},  # 126KB (plus overhead), ok
            ]
        )

    s = str(exc_info.value)
    assert s.count("DocumentTooLarge") == 2
    assert 'doc_id: "one"' in s
    assert 'doc_id: "three"' in s
    assert 'doc_id: "two"' not in s
    assert 'doc_id: "four"' not in s
    assert "max_size_bytes: 131072" in s


@pytest.mark.parametrize(
    "params",
    [
        (True, schema.bool()),
        ("hello", schema.text()),
        (1, schema.int()),
        (1.0, schema.float()),
        (b"hello", schema.bytes()),
    ],
)
def test_upsert_primitives(ctx: ProjectContext, params):
    value, data_type = params  # type: ignore

    collection = ctx.client.collections().create(
        ctx.scope("test"),
        schema={"field": data_type},  # type: ignore
    )

    lsn = ctx.client.collection(collection.name).upsert(
        [
            {"_id": "x", "field": value},
        ]
    )

    obj = ctx.client.collection(collection.name).get(["x"], lsn=lsn)

    assert obj["x"]["field"] == value


def test_upsert_with_bytes_helper(ctx: ProjectContext):
    collection = ctx.client.collections().create(
        ctx.scope("test-bytes"),
        schema={
            "title": schema.text().required(),
            "thumbnail": schema.bytes(),
        },
    )

    lsn = ctx.client.collection(collection.name).upsert(
        [
            {
                "_id": "doc1",
                "rank": 1,
                "title": "Document with bytes from list",
                "thumbnail": data.bytes([0, 1, 255, 128]),
            },
            {
                "_id": "doc2",
                "rank": 2,
                "title": "Document with bytes from bytes object",
                "thumbnail": data.bytes(b"\x00\x01\xff\x80"),
            },
            {
                "_id": "doc3",
                "rank": 3,
                "title": "Document with empty bytes",
                "thumbnail": data.bytes([]),
            },
            {
                "_id": "doc4",
                "rank": 4,
                "title": "Document with native bytes",
                "thumbnail": bytes([10, 20, 30]),  # Test native Python bytes still work
            },
        ]
    )

    ctx.client.collection(collection.name).count(lsn=lsn)

    results = ctx.client.collection(collection.name).query(
        query.select("title", "thumbnail").topk(query.field("rank"), 10, True)
    )

    assert len(results) == 4

    for doc in results:
        if doc["_id"] == "doc1":
            assert doc["thumbnail"] == bytes([0, 1, 255, 128])
        elif doc["_id"] == "doc2":
            assert doc["thumbnail"] == bytes([0, 1, 255, 128])
        elif doc["_id"] == "doc3":
            assert doc["thumbnail"] == bytes([])
        elif doc["_id"] == "doc4":
            assert doc["thumbnail"] == bytes([10, 20, 30])


def test_upsert_vectors(ctx: ProjectContext):
    collection = ctx.client.collections().create(
        ctx.scope("test"),
        schema={
            "f32_vector": schema.f32_vector(3),
            "u8_vector": schema.u8_vector(3),
            "binary_vector": schema.binary_vector(3),
        },
    )

    lsn = ctx.client.collection(collection.name).upsert(
        [
            {
                "_id": "x",
                "f32_vector": [1, 2, 3],
                "u8_vector": data.u8_vector([4, 5, 6]),
                "binary_vector": data.binary_vector([7, 8, 9]),
            }
        ]
    )

    obj = ctx.client.collection(collection.name).get(["x"], lsn=lsn)

    assert obj["x"]["f32_vector"] == [1, 2, 3]
    assert obj["x"]["u8_vector"] == [4, 5, 6]
    assert obj["x"]["binary_vector"] == [7, 8, 9]


def test_upsert_sparse_vectors(ctx: ProjectContext):
    collection = ctx.client.collections().create(
        ctx.scope("test"),
        schema={
            "f32_sparse_vector": schema.f32_sparse_vector(),
            "u8_sparse_vector": schema.u8_sparse_vector(),
        },
    )

    lsn = ctx.client.collection(collection.name).upsert(
        [
            {
                "_id": "x",
                "f32_sparse_vector": {1: 1.2, 2: 2.3, 3: 3.4},
                "u8_sparse_vector": data.u8_sparse_vector({1: 4, 2: 5, 3: 6}),
            }
        ]
    )

    obj = ctx.client.collection(collection.name).get(["x"], lsn=lsn)

    assert set(obj["x"]["f32_sparse_vector"].keys()) == {1, 2, 3}
    assert "{:.2f}".format(obj["x"]["f32_sparse_vector"][1]) == "1.20"
    assert "{:.2f}".format(obj["x"]["f32_sparse_vector"][2]) == "2.30"
    assert "{:.2f}".format(obj["x"]["f32_sparse_vector"][3]) == "3.40"
    assert set(obj["x"]["u8_sparse_vector"].keys()) == {1, 2, 3}
    assert obj["x"]["u8_sparse_vector"][1] == 4
    assert obj["x"]["u8_sparse_vector"][2] == 5
    assert obj["x"]["u8_sparse_vector"][3] == 6

def test_upsert_empty_float_list(ctx: ProjectContext):
    collection = ctx.client.collections().create(
        ctx.scope("test"),
        schema={"f32_list": schema.list(value_type="float")},
    )

    lsn = ctx.client.collection(collection.name).upsert(
        [{"_id": "x", "f32_list": []}]
    )

    obj = ctx.client.collection(collection.name).get(["x"], lsn=lsn)

    assert obj["x"]["f32_list"] == []


def test_upsert_empty_float_list_with_helper(ctx: ProjectContext):
    collection = ctx.client.collections().create(
        ctx.scope("test"),
        schema={"f32_list": schema.list(value_type="float")},
    )

    lsn = ctx.client.collection(collection.name).upsert(
        [{"_id": "x", "f32_list": data.f32_list([])}]
    )

    obj = ctx.client.collection(collection.name).get(["x"], lsn=lsn)

    assert obj["x"]["f32_list"] == []


def test_upsert_empty_integer_list_raises_error(ctx: ProjectContext):
    collection = ctx.client.collections().create(
        ctx.scope("test"),
        schema={"i32_list": schema.list(value_type="integer")},
    )

    with pytest.raises(error.DocumentValidationError) as exc_info:
        ctx.client.collection(collection.name).upsert(
            [{"_id": "x", "i32_list": []}]
        )
    assert "field: \"i32_list\", expected_type: \"list<integer>\", got_value: \"list<f32>\"" in str(exc_info.value)


def test_upsert_empty_integer_list_with_helper(ctx: ProjectContext):

    collection = ctx.client.collections().create(
        ctx.scope("test"),
        schema={"i32_list": schema.list(value_type="integer")},
    )

    lsn = ctx.client.collection(collection.name).upsert(
        [{"_id": "x", "i32_list": data.i32_list([])}]
    )

    obj = ctx.client.collection(collection.name).get(["x"], lsn=lsn)

    assert obj["x"]["i32_list"] == []


def test_upsert_empty_string_list_raises_error(ctx: ProjectContext):
    collection = ctx.client.collections().create(
        ctx.scope("test"),
        schema={"string_list": schema.list(value_type="text")},
    )

    with pytest.raises(error.DocumentValidationError) as exc_info:
        ctx.client.collection(collection.name).upsert(
        [{"_id": "x", "string_list": []}]
    )

    assert "field: \"string_list\", expected_type: \"list<string>\", got_value: \"list<f32>\"" in str(exc_info.value)


def test_upsert_empty_string_list_with_helper(ctx: ProjectContext):
    collection = ctx.client.collections().create(
        ctx.scope("test"),
        schema={"string_list": schema.list(value_type="text")},
    )

    lsn = ctx.client.collection(collection.name).upsert(
        [{"_id": "x", "string_list": data.string_list([])}]
    )

    obj = ctx.client.collection(collection.name).get(["x"], lsn=lsn)

    assert obj["x"]["string_list"] == []



import pytest
from topk_sdk import error
from topk_sdk.query import field, filter, fn, not_, select, literal, match, min, max, abs

from . import ProjectContext
from .utils import dataset, doc_ids


def test_query_lte(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(field("published_year") <= 1950).topk(
            field("published_year"), 100, True
        )
    )

    assert doc_ids(result) == {"1984", "pride", "hobbit", "moby", "gatsby"}


def test_query_and(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(
            (field("published_year") <= 1950) & (field("published_year") >= 1948)
        ).topk(field("published_year"), 100, True)
    )

    assert doc_ids(result) == {"1984"}

def test_query_is_null(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(field("nullable_embedding").is_null()).topk(field("published_year"), 100, True)
    )

    assert doc_ids(result) == {"pride", "gatsby", "moby", "hobbit", "lotr", "alchemist"}

def test_query_is_not_null(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(field("nullable_embedding").is_not_null()).topk(field("published_year"), 100, True)
    )

    assert doc_ids(result) == {"mockingbird", "1984", "catcher", "harry"}

def test_query_not(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(not_(field("_id").contains("gatsby"))).topk(
            field("published_year"), 100, False
        )
    )

    assert doc_ids(result) == {
        "harry",
        "lotr",
        "1984",
        "mockingbird",
        "moby",
        "alchemist",
        "catcher",
        "hobbit",
        "pride",
    }

def test_query_choose_literal(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            love_score=field("summary")
            .match_all("love")
            .choose(literal(2.0), literal(0.1))
        )
        .filter(field("love_score") > 1.0)
        .topk(field("love_score"), 10, False)
    )

    assert doc_ids(result) == {"pride", "gatsby"}

def test_query_choose_literal_and_field(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            love_score=field("summary")
            .match_all("love")
            .choose(field("published_year"), literal(10))
        )
        .topk(field("love_score"), 2, False)
    )

    assert result == [
        {"_id": "gatsby", "love_score": 1925},
        {"_id": "pride", "love_score": 1813},
    ]

def test_query_choose_field(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            love_score=field("summary")
            .match_all("love")
            .choose(field("published_year"), field("published_year") / 10)
        )
        .topk(field("love_score"), 3, False)
    )

    assert result == [
        {"_id": "gatsby", "love_score": 1925},
        {"_id": "pride", "love_score": 1813},
        {"_id": "harry", "love_score": 199},
    ]

def test_query_coalesce_nullable(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(importance=field("nullable_importance").coalesce(1.0))
        .filter(field("published_year") < 1900)
        .topk(field("published_year"), 3, False)
    )

    assert result == [
        {"_id": "moby", "importance": 5.0},
        {"_id": "pride", "importance": 1.0},
    ]

def test_query_coalesce_missing(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(importance=field("missing_field").coalesce(1.0))
        .filter(field("published_year") < 1900)
        .topk(field("published_year"), 3, False)
    )

    assert result == [
        {"_id": "moby", "importance": 1.0},
        {"_id": "pride", "importance": 1.0},
    ]

def test_query_coalesce_non_nullable(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(coalesced_year=field("published_year").coalesce(0))
        .filter(field("published_year") < 1900)
        .topk(field("published_year"), 3, False)
    )

    assert result == [
        {"_id": "moby", "coalesced_year": 1851},
        {"_id": "pride", "coalesced_year": 1813},
    ]

def test_query_abs(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(abs_year=abs(field("published_year") - 1990)).topk(
            field("abs_year"), 3, True
        )
    )

    # The 3 books closest to 1990
    assert result == [
        {"_id": "alchemist", "abs_year": 2},
        {"_id": "harry", "abs_year": 7},
        {"_id": "mockingbird", "abs_year": 30},
    ]

def test_query_topk_min_max(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(bm25_score=fn.bm25_score())
        .select(clamped_bm25_score=max(min(field("bm25_score"), 2.0), 1.6))
        .filter(match("millionaire love consequences dwarves", field="summary", weight=1.0, all=False))
        .topk(field("clamped_bm25_score"), 5, False)
    )

    assert len(result) == 4

    assert result[0]["_id"] == "gatsby"
    assert result[0]["clamped_bm25_score"] == 2.0

    assert result[1]["_id"] == "hobbit"
    clamped_score_1 = result[1]["clamped_bm25_score"]
    assert 1.6 <= clamped_score_1 <= 2.0

    assert result[2]["_id"] == "moby"
    clamped_score_2 = result[2]["clamped_bm25_score"]
    assert 1.6 <= clamped_score_2 <= 2.0

    assert result[3]["_id"] == "pride"
    assert result[3]["clamped_bm25_score"] == 1.6

def test_query_gt_and_lte_string(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(field("_id").gt("moby") & (field("_id").lte("pride"))).topk(
            field("published_year"), 100, True
        )
    )

    assert doc_ids(result) == {"mockingbird", "pride"}

def test_query_min_string(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select("title", min_string=field("title").min("Oz"))
        .topk(field("published_year"), 2, True)
    )

    assert result == [
        {"_id": "pride", "title": "Pride and Prejudice", "min_string": "Oz"},
        {"_id": "moby", "title": "Moby-Dick", "min_string": "Moby-Dick"},
    ]



import math
import pytest
from topk_sdk import error
from topk_sdk.query import field, filter, select, fn, match

from . import ProjectContext
from .utils import dataset, doc_ids, doc_ids_ordered


def test_query_exp_ln(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(bm25_score=fn.bm25_score())
        .select(
            bm25_score_scale=(field("bm25_score") * 1.5).exp(),
            bm25_score_smooth=(field("bm25_score") + 1).ln(),
        )
        .filter(match("millionaire love consequences dwarves", field="summary", weight=1.0, all=False))
        .topk(field("bm25_score_scale"), 2, False)
    )

    assert doc_ids_ordered(result) == ["gatsby", "hobbit"]

    for doc in result:
        bm25_score = doc["bm25_score"]
        bm25_score_scale = doc["bm25_score_scale"]
        bm25_score_smooth = doc["bm25_score_smooth"]
        assert abs(math.exp(bm25_score * 1.5) - bm25_score_scale) < 1e-4
        assert abs(math.log(bm25_score + 1.0) - bm25_score_smooth) < 1e-4


def test_query_float_inf(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(to_infinity=field("published_year").exp()).topk(
            field("published_year"), 2, True
        )
    )

    assert len(result) == 2

    for doc in result:
        to_infinity = doc["to_infinity"]
        assert to_infinity == float("inf")


def test_query_sqrt_square(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            published_year=field("published_year"),
            published_year_2=field("published_year").sqrt().square(),
        ).topk(field("published_year_2"), 2, True)
    )

    assert doc_ids_ordered(result) == ["pride", "moby"]

    for doc in result:
        year_2 = doc["published_year_2"]
        year_orig = doc["published_year"]
        assert round(year_2) == year_orig


def test_query_sqrt_filter(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(title=field("title"))
        .filter(field("published_year").sqrt() > math.sqrt(1990))
        .topk(field("published_year"), 2, True)
    )

    assert result == [{"_id": "harry", "title": "Harry Potter and the Sorcerer's Stone"}]



from topk_sdk.query import field, fn, select, match

from . import ProjectContext
from .utils import dataset, doc_ids_ordered, is_sorted


def test_query_hybrid_vector_bm25(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            summary_distance=fn.vector_distance("summary_embedding", [2.0] * 16),
            bm25_score=fn.bm25_score(),
        )
        .filter(
            match("love", None, 30.0, False) | (match("young", None, 10.0, False))
        )
        .topk(
            field("bm25_score") + (field("summary_distance") * 100), 2, True
        )
    )

    assert len(result) == 2
    assert doc_ids_ordered(result) == ["mockingbird", "pride"]


def test_query_hybrid_keyword_boost(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    # Multiply summary_distance by 0.1 if the summary matches "racial injustice", otherwise
    # multiply by 1.0 (leave unchanged).
    for score_expr in [
        field("summary_distance")
        * (field("summary").match_all("racial injustice").choose(0.1, 1.0)),
        field("summary_distance").boost(field("summary").match_all("racial injustice"), 0.1),
    ]:
        result = ctx.client.collection(collection.name).query(
            select(
                summary_distance=fn.vector_distance("summary_embedding", [2.3] * 16)
            )
            .topk(score_expr, 3, True)
        )

        # Keyword boosting swaps the order of results so we expect [1984, mockingbird, pride]
        # instead of [1984, pride, mockingbird].
        assert doc_ids_ordered(result) == ["1984", "mockingbird", "pride"]

        # We use a modified scoring expression so the results are not sorted by summary_distance.
        assert not is_sorted(result, "summary_distance")

def test_boost_helper_same_expression():
    no_helper = field("summary_distance") * (field("summary").match_all("racial injustice").choose(0.1, 1.0))
    with_helper = field("summary_distance").boost(field("summary").match_all("racial injustice"), 0.1)
    assert no_helper == with_helper

def test_query_hybrid_coalesce_score(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            summary_score=fn.vector_distance("summary_embedding", [4.1] * 16),
            nullable_score=fn.vector_distance("nullable_embedding", [4.1] * 16),
        )
        .topk(
            field("summary_score") + field("nullable_score").coalesce(0.0), 3, True
        )
    )

    # Adding the nullable_score without coalescing would exclude "pride" and "gatsby" from
    # the result set, even though they are the closest candidates based on summary_score.
    assert doc_ids_ordered(result) == ["gatsby", "pride", "catcher"]



import pytest
from topk_sdk.data import f32_list, f64_list, i32_list, i64_list, string_list, u32_list


def test_u32_list():
    result = u32_list([0, 1, 255, 4294967295])
    assert result is not None
    assert str(result) == "List(U32([0, 1, 255, 4294967295]))"


def test_u32_list_empty():
    result = u32_list([])
    assert result is not None
    assert str(result) == "List(U32([]))"


def test_u32_list_invalid_input():
    with pytest.raises(
        TypeError, match="Expected list\\[int\\] for u32_list\\(\\) function"
    ):
        u32_list("not a list")  # type: ignore

    with pytest.raises(
        TypeError, match="Expected list\\[int\\] for u32_list\\(\\) function"
    ):
        u32_list(123)  # type: ignore


def test_i32_list():
    result = i32_list([-2147483648, -1, 0, 1, 2147483647])
    assert result is not None
    assert str(result) == "List(I32([-2147483648, -1, 0, 1, 2147483647]))"


def test_i32_list_empty():
    result = i32_list([])
    assert result is not None
    assert str(result) == "List(I32([]))"


def test_i32_list_invalid_input():
    with pytest.raises(
        TypeError, match="Expected list\\[int\\] for i32_list\\(\\) function"
    ):
        i32_list("not a list")  # type: ignore

    with pytest.raises(
        TypeError, match="Expected list\\[int\\] for i32_list\\(\\) function"
    ):
        i32_list(123)  # type: ignore


def test_i64_list():
    result = i64_list([-9223372036854775808, -1, 0, 1, 9223372036854775807])
    assert result is not None
    assert str(result) == "List(I64([-9223372036854775808, -1, 0, 1, 9223372036854775807]))"


def test_i64_list_empty():
    result = i64_list([])
    assert result is not None
    assert str(result) == "List(I64([]))"


def test_i64_list_invalid_input():
    with pytest.raises(
        TypeError, match="Expected list\\[int\\] for i64_list\\(\\) function"
    ):
        i64_list("not a list")  # type: ignore

    with pytest.raises(
        TypeError, match="Expected list\\[int\\] for i64_list\\(\\) function"
    ):
        i64_list(123)  # type: ignore


def test_f32_list():
    result = f32_list([1.0, 2.5, -3.5, 0.0])
    assert result is not None
    assert str(result) == "List(F32([1.0, 2.5, -3.5, 0.0]))"



def test_f32_list_empty():
    result = f32_list([])
    assert result is not None
    assert str(result) == "List(F32([]))"


def test_f32_list_invalid_input():
    with pytest.raises(
        TypeError, match="Expected list\\[float\\] for f32_list\\(\\) function"
    ):
        f32_list("not a list")  # type: ignore

    with pytest.raises(
        TypeError, match="Expected list\\[float\\] for f32_list\\(\\) function"
    ):
        f32_list(123)  # type: ignore


def test_f64_list():
    result = f64_list([1.0, 2.5, -3.14, 0.0])
    assert result is not None
    assert str(result) == "List(F64([1.0, 2.5, -3.14, 0.0]))"


def test_f64_list_empty():
    result = f64_list([])
    assert result is not None
    assert str(result) == "List(F64([]))"


def test_f64_list_invalid_input():
    with pytest.raises(
        TypeError, match="Expected list\\[float\\] for f64_list\\(\\) function"
    ):
        f64_list("not a list")  # type: ignore

    with pytest.raises(
        TypeError, match="Expected list\\[float\\] for f64_list\\(\\) function"
    ):
        f64_list(123)  # type: ignore


def test_string_list():
    result = string_list(["1", "2", "3"])
    assert result is not None
    assert str(result) == "List(String([\"1\", \"2\", \"3\"]))"


def test_string_list_empty():
    result = string_list([])
    assert result is not None
    assert str(result) == "List(String([]))"

def test_string_list_invalid_input():
    with pytest.raises(
        TypeError, match="Expected list\\[str\\] for string_list\\(\\) function"
    ):
        string_list("not a list")  # type: ignore

    with pytest.raises(
        TypeError, match="Expected list\\[str\\] for string_list\\(\\) function"
    ):
        string_list(123)  # type: ignore



import asyncio
import pytest
from topk_sdk.query import select, field, literal

from . import AsyncProjectContext
from .utils import dataset, doc_ids


@pytest.mark.asyncio
async def test_async_upsert(async_ctx: AsyncProjectContext):
    collection = await async_ctx.client.collections().create(async_ctx.scope("test"), schema={})
    async_collection = async_ctx.client.collection(collection.name)

    lsn = await async_collection.upsert([{"_id": "one"}])
    assert lsn == "1"


@pytest.mark.asyncio
async def test_async_get(async_ctx: AsyncProjectContext):
    collection = await dataset.books.setup_async(async_ctx)
    async_collection = async_ctx.client.collection(collection.name)

    docs = await async_collection.get(["lotr"])

    assert docs == {
        "lotr": {
            "_id": "lotr",
            "title": "The Lord of the Rings: The Fellowship of the Ring",
            "published_year": 1954,
            "summary": "A group of unlikely heroes sets out to destroy a powerful, evil ring.",
            "summary_embedding": [9.0] * 16,
            "sparse_f32_embedding": {9: 1.0, 10: 2.0, 11: 3.0},
            "sparse_u8_embedding": {9: 1, 10: 2, 11: 3},
            "tags": ["lord of the rings", "fellowship", "magic", "wizard", "elves"],
            "codes": ["ISBN 978-0-547-92821-0", "ISBN 0-547-92821-2", "OCLC 434394005", "LCCN 2004558654", "Barcode 0618346252"],
            "user_ratings": ["epic", "legendary", "good"],
        }
    }


@pytest.mark.asyncio
async def test_async_delete(async_ctx: AsyncProjectContext):
    collection = await async_ctx.client.collections().create(async_ctx.scope("test"), schema={})
    async_collection = async_ctx.client.collection(collection.name)

    lsn = await async_collection.upsert(
        [
            {"_id": "one", "rank": 1},
            {"_id": "two", "rank": 2},
        ]
    )
    assert lsn == "1"

    # wait for write to be flushed
    await async_collection.count()

    lsn = await async_collection.delete(["one"])
    assert lsn == "2"

    docs = await async_collection.query(
        select("title").topk(field("rank"), 100, True), lsn=lsn
    )

    assert doc_ids(docs) == {"two"}


@pytest.mark.asyncio
async def test_async_query(async_ctx: AsyncProjectContext):
    collection = await dataset.books.setup_async(async_ctx)

    results = await async_ctx.client.collection(collection.name).query(
        select("title", "published_year", literal=literal(1.0))
        .filter(field("title") == "1984")
        .topk(field("published_year"), 100, True)
    )

    assert results == [{"_id": "1984", "title": "1984", "published_year": 1949, "literal": 1.0}]


test_documents = [
    {
        "_id": "test_doc",
        "title": "Test Document",
        "published_year": 2023,
        "summary": "A test document for async collection operations.",
        "summary_embedding": [0.5] * 16,
    }
]


@pytest.mark.asyncio
async def test_async_collection_parallel_queries(async_ctx: AsyncProjectContext):
    collection = await dataset.books.setup_async(async_ctx)
    async_collection = async_ctx.client.collection(collection.name)

    # Upsert test documents
    lsn = await async_collection.upsert(test_documents)

    # Define multiple queries
    query1 = select("title", "published_year").topk(field("published_year"), 5)
    query2 = select("title", "summary").topk(field("published_year"), 3)
    query3 = select("published_year").topk(field("published_year"), 10)

    # Execute queries in parallel
    results = await asyncio.gather(
        async_collection.query(query1, lsn=lsn),
        async_collection.query(query2, lsn=lsn),
        async_collection.query(query3, lsn=lsn)
    )

    # Verify all queries returned results
    assert len(results) == 3
    assert all(len(result) >= 1 for result in results)

    # Verify query-specific fields are present
    assert "title" in results[0][0]
    assert "published_year" in results[0][0]

    assert "title" in results[1][0]
    assert "summary" in results[1][0]

    assert "published_year" in results[2][0]


def test_async_collection_sync_usage(async_ctx: AsyncProjectContext):
    async def async_operations():
        collection = await dataset.books.setup_async(async_ctx)
        async_collection = async_ctx.client.collection(collection.name)

        # Perform async operations
        lsn = await async_collection.upsert(test_documents)

        docs = await async_collection.get(["test_doc"])
        assert "test_doc" in docs

        return lsn

    # Run async operations in sync context
    lsn = asyncio.run(async_operations())
    assert isinstance(lsn, str)



import pytest
from topk_sdk import error

from . import AsyncProjectContext


@pytest.mark.asyncio
async def test_async_list_collections(async_ctx: AsyncProjectContext):
    c = await async_ctx.client.collections().create(async_ctx.scope("test"), schema={})
    response = await async_ctx.client.collections().list()
    assert c in response


@pytest.mark.asyncio
async def test_async_create_collection(async_ctx: AsyncProjectContext):
    c = await async_ctx.client.collections().create(async_ctx.scope("test"), schema={})
    collections = await async_ctx.client.collections().list()
    assert c in collections


@pytest.mark.asyncio
async def test_async_create_duplicate_collection(async_ctx: AsyncProjectContext):
    await async_ctx.client.collections().create(async_ctx.scope("test"), schema={})

    with pytest.raises(error.CollectionAlreadyExistsError):
        await async_ctx.client.collections().create(async_ctx.scope("test"), schema={})


@pytest.mark.asyncio
async def test_async_delete_non_existent_collection(async_ctx: AsyncProjectContext):
    with pytest.raises(error.CollectionNotFoundError):
        await async_ctx.client.collections().delete(async_ctx.scope("test"))


@pytest.mark.asyncio
async def test_async_delete_collection(async_ctx: AsyncProjectContext):
    c = await async_ctx.client.collections().create(async_ctx.scope("test"), schema={})
    await async_ctx.client.collections().delete(async_ctx.scope("test"))

    collections = await async_ctx.client.collections().list()
    assert c not in collections


@pytest.mark.asyncio
async def test_async_get_collection(async_ctx: AsyncProjectContext):
    # Test getting non-existent collection
    with pytest.raises(error.CollectionNotFoundError):
        await async_ctx.client.collections().get(async_ctx.scope("test"))

    # Create collection
    c = await async_ctx.client.collections().create(async_ctx.scope("test"), schema={})

    # Get collection
    collection = await async_ctx.client.collections().get(async_ctx.scope("test"))
    assert collection == c



from typing import Any


def doc_ids(result: list[dict[str, Any]]) -> set[str]:
    return {doc["_id"] for doc in result}

def doc_ids_ordered(result: list[dict[str, Any]]) -> list[str]:
    return [doc["_id"] for doc in result]

def is_sorted(result: list[dict[str, Any]], field_name: str) -> bool:
    values = [doc[field_name] for doc in result]
    return all(values[i] <= values[i + 1] for i in range(len(values) - 1))
